-- $Id$
Broker_Friends = LibStub("AceAddon-3.0"):NewAddon("Broker_Friends", "AceEvent-3.0", "AceConsole-3.0", "AceTimer-3.0")
local addon = Broker_Friends

local L = LibStub("AceLocale-3.0"):GetLocale("Broker_Friends")

-- Grab used libraries.
local broker = LibStub("LibDataBroker-1.1")
local qtip = LibStub("LibQTip-1.0")

-- Create the data feed.
local feed = broker:NewDataObject("Broker_Friends", {
	type = "data source",
	text = "",
	icon = "Interface\\FriendsFrame\\FriendsFrameScrollIcon",

	OnEnter = function(anchor)
		addon:ShowTooltipAt(anchor)
	end,

	OnClick = function()
		ToggleFriendsFrame(1)
	end,
})

-- QTip tooltip holder.
local tooltip

-- Realm & faction data used for comparison. Filled OnEnable.
local playerRealm, playerFaction

-- Upvalues
local GetNumFriends, GetFriendInfo = GetNumFriends, GetFriendInfo
local BNFeaturesEnabledAndConnected, BNGetNumFriends, BNGetFriendInfo, BNGetNumFriendToons, BNGetFriendToonInfo =
	BNFeaturesEnabledAndConnected, BNGetNumFriends, BNGetFriendInfo, BNGetNumFriendToons, BNGetFriendToonInfo

-- Blizzard returns localized class names in the friends API. This is a huge pain, so we'll just make a list
-- of localized class names to class color codes here once for easy lookup.
local localizedClassColorTable = {}
for key, localized in pairs(LOCALIZED_CLASS_NAMES_MALE) do
	local color = RAID_CLASS_COLORS[key]
	localizedClassColorTable[localized] = ("|cff%02x%02x%02x"):format(color.r * 255, color.g * 255, color.b * 255)
end
for key, localized in pairs(LOCALIZED_CLASS_NAMES_FEMALE) do
	local color = RAID_CLASS_COLORS[key]
	localizedClassColorTable[localized] = ("|cff%02x%02x%02x"):format(color.r * 255, color.g * 255, color.b * 255)
end

function addon:OnEnable()
	-- Register basic updates on three events that FriendsFrame updates on.
	-- FriendsFrame also updates on PARTY_MEMBERS_CHANGED for the RaF features but that doesn't interest us.
	self:RegisterEvent("FRIENDLIST_UPDATE", "Update")
	self:RegisterEvent("BN_FRIEND_LIST_SIZE_CHANGED", "Update")
	self:RegisterEvent("BN_FRIEND_INFO_CHANGED", "Update")
	
	-- Make sure we have data and keep it somewhat up-to-date. The game is pretty decent with keeping it
	-- updated itself, so once a minute is more than enough.
	ShowFriends()
	self:ScheduleRepeatingTimer(ShowFriends, 60)
	
	-- Initialize realm/faction data, it should be ok now.
	playerRealm = GetRealmName()
	playerFaction = UnitFactionGroup("player")
end

-- Update the text display and possibly the tooltip.
function addon:Update()
	-- Basic friend count
	local normalFriends, onlineNormalFriends = GetNumFriends()
	
	local BNfriends, onlineBNfriends = 0, 0
	if BNFeaturesEnabledAndConnected() then
		BNfriends, onlineBNfriends = BNGetNumFriends()
	end
	
	local totalOnline = onlineNormalFriends + onlineBNfriends
	local totalAll = normalFriends + BNfriends
	feed.text = ("%d/%d"):format(totalOnline, totalAll)
	
	-- If we have a tooltip, great! Update it too, then.
	if qtip:IsAcquired(self) then
		self:UpdateTooltip()
	end
end

local function charNameClicked(_, charName)
	SetItemRef(("player:%s"):format(charName), ("|Hplayer:%s|h[%s]|h"):format(charName, charName), "LeftButton")
end

-- QTip only allows you to pass one argument, so we pass the presenceID and look up the real name ourselves.
-- This saves creating an amount of tables equal to your RealID friends every update (!) at the cost of
-- another function call when we click on a RealID friend.
local function bnetNameClicked(_, presenceID)
	local _, givenName, surname = BNGetFriendInfoByID(presenceID)
	local name = BATTLENET_NAME_FORMAT:format(givenName, surname)
	local itemRef = ("%s:%d"):format(name, presenceID)
	SetItemRef(("BNplayer:%s"):format(itemRef), ("|HBNplayer:%s|h[%s]|h"):format(itemRef, name), "LeftButton")
end

-- Update the tooltip.
function addon:UpdateTooltip()
	tooltip:Clear()

	tooltip:SetCell(tooltip:AddLine(), 1, "Broker_Friends", GameFontNormalMed3, nil, 4)

	tooltip:AddHeader(L["Name"], L["Level"], L["Area"], L["Note"])
	local normalAll, normalOnline = GetNumFriends()
	for i = 1, normalOnline do
		local name, level, class, area, connected, status, note = GetFriendInfo(i)
		
		if connected then
			local displayedName = name
			
			if status then
				displayedName = ("%s%s"):format(status, displayedName)
			end
			
			if localizedClassColorTable[class] then
				displayedName = ("%s%s%s"):format(localizedClassColorTable[class], displayedName, "|r")
			end
			
			local line = tooltip:AddLine(displayedName, level, area, note)
			tooltip:SetCellScript(line, 1, "OnMouseDown", charNameClicked, name)
		end
	end
	
	if BNFeaturesEnabledAndConnected() then
		local bnAll, bnOnline = BNGetNumFriends()
		-- Add an empty line between normal and RealID friends if you have both types of friends online.
		if normalOnline > 0 and bnOnline > 0 then
			tooltip:AddLine(" ")
		end
		
		for friend = 1, bnOnline do
			local presenceID, givenName, surname, bnCharName, _, _, isOnline, _,
				isAFK, isDND, _, noteText, _, _ = BNGetFriendInfo(friend)
			local fullName = BATTLENET_NAME_FORMAT:format(givenName, surname)
			local playerStatus = ""
			if isAFK then
				playerStatus = ("<%s>"):format(AFK)
			elseif isDND then
				playerStatus = ("<%s>"):format(DND)
			end
			
			if isOnline then
				-- Is a friend playing WoW? If so, we add his WoW toon(s) - if not, we add his RealID/SC2 info below.
				-- If a friend is playing WoW and SC2, this below code should only display his WoW toon - that's intended.
				local inWow = false
				local lastClient = "UNKNOWN"
				local numToons = BNGetNumFriendToons(friend)
				
				-- TODO: Test what happens when multiple toons are active.
				for toon = 1, numToons do
					local _, charName, gameClient, charRealm, factionNum, _, charClass, _,
						charArea, charLevel, gameText = BNGetFriendToonInfo(friend, toon)
					
					lastClient = gameClient
					
					-- *If* he's playing WoW, we add a proper toon line here.
					if gameClient == BNET_CLIENT_WOW then
						inWow = true
						
						-- Ok so, the API gives us back a faction number which we can turn into a real faction string
						-- by looking it up in this table.
						local actualFaction = PLAYER_FACTION_GROUP[factionNum]

						-- Blizzard provides an API for this. Guess what? It uses the toonid which we don't have.
						-- Even the actual friends frame does it this way. We shouldn't have to do this...
						local canCooperate = false
						local displayedCharName = charName
						if charRealm == playerRealm and actualFaction == playerFaction then
							canCooperate = true
						else
							displayedCharName = ("%s-%s*"):format(charName, charRealm)
						end
						
						-- Status stuff.
						displayedCharName = playerStatus .. displayedCharName
						-- Color crap.
						if localizedClassColorTable[charClass] then
							displayedCharName = ("%s%s|r"):format(localizedClassColorTable[charClass], displayedCharName)
						end
						
						-- Finally, we get to add the line.
						local line = tooltip:AddLine(displayedCharName, charLevel, charArea, fullName)
						
						-- Set up whisper on click if we can "cooperate" (same faction/realm).
						if canCooperate then
							tooltip:SetCellScript(line, 1, "OnMouseDown", charNameClicked, charName)
						end

						-- Also set up B.net whispers in the note slot.
						tooltip:SetCellScript(line, 4, "OnMouseDown", bnetNameClicked, presenceID)
					end
				end
				
				-- If this friend isn't playing WoW at all, we add a different line here.
				-- TODO: Test this with various combinations to make sure it works as intended.
				if not inWow then
					local displayedCharName = bnCharName
					displayedCharName = ("%s%s%s|r"):format(FRIENDS_OTHER_NAME_COLOR_CODE, playerStatus, displayedCharName)
					tooltip:AddLine(bnCharName, "-", (L["In-game: %s"]):format(lastClient), fullName)
					tooltip:SetCellScript(line, 4, "OnMouseDown", bnetNameClicked, presenceID)
				end
			end -- end online check
		end -- end bnet iterator
	elseif BNFeaturesEnabled() and not BNConnected() then
		-- Might as well use the text from GlobalStrings
		tooltip:SetCell(tooltip:AddLine(), 1, BATTLENET_UNAVAILABLE, nil, 4)
	end -- end bnet conditional
	
	-- Use the full colspan
	tooltip:SetCell(tooltip:AddLine(), 1, L["|cffeda55fClick|r on the name or real name of a friend to send a whisper."], nil, 4)
end

function addon:ShowTooltipAt(anchor)
	ShowFriends()
	tooltip = qtip:Acquire(Broker_Friends, 4, "LEFT", "CENTER", "CENTER", "RIGHT")
	tooltip:SetHeaderFont(GameFontNormal)
	tooltip:SetAutoHideDelay(0.2, anchor)
	tooltip:SmartAnchorTo(anchor)
	self:UpdateTooltip()
	tooltip:Show()
end
